/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.nifi.wali;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;

public class BlockingQueuePool<T> implements ObjectPool<T> {
    private final BlockingQueue<T> queue;
    private final Supplier<T> creationFunction;
    private final Predicate<T> reuseCheck;
    private final Consumer<T> returnPreparation;

    public BlockingQueuePool(final int maxSize, final Supplier<T> creationFunction, final Predicate<T> reuseCheck, final Consumer<T> returnPreparation) {
        this.queue = new LinkedBlockingQueue<>(maxSize);
        this.creationFunction = creationFunction;
        this.reuseCheck = reuseCheck;
        this.returnPreparation = returnPreparation;
    }

    @Override
    public T borrowObject() {
        final T existing = queue.poll();
        if (existing != null) {
            return existing;
        }

        return creationFunction.get();
    }

    @Override
    public void returnObject(final T somethingBorrowed) {
        if (reuseCheck.test(somethingBorrowed)) {
            returnPreparation.accept(somethingBorrowed);
            queue.offer(somethingBorrowed);
        }
    }
}
