/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.nifi.processors.opentelemetry.encoding;

import com.fasterxml.jackson.databind.BeanDescription;
import com.fasterxml.jackson.databind.DeserializationConfig;
import com.fasterxml.jackson.databind.JavaType;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.google.protobuf.Message;
import com.hubspot.jackson.datatype.protobuf.MessageDeserializerFactory;
import com.hubspot.jackson.datatype.protobuf.ProtobufJacksonConfig;

/**
 * Standard extension of Deserializer Factory supporting hexadecimal decoding to ByteString
 */
public class StandardMessageDeserializerFactory extends MessageDeserializerFactory {
    private final ProtobufJacksonConfig protobufJacksonConfig;

    public StandardMessageDeserializerFactory(final ProtobufJacksonConfig protobufJacksonConfig) {
        super(protobufJacksonConfig);
        this.protobufJacksonConfig = protobufJacksonConfig;
    }

    @Override
    @SuppressWarnings("unchecked")
    public JsonDeserializer<?> findBeanDeserializer(final JavaType type, final DeserializationConfig config, final BeanDescription beanDescription) throws JsonMappingException {

        final Class<?> rawClass = type.getRawClass();
        if (isHexadecimalMessageType(rawClass)) {
            final Class<? extends Message> messageClass = (Class<? extends Message>) rawClass;
            return new ByteStringFieldDeserializer<>(messageClass, protobufJacksonConfig).buildAtEnd();
        } else {
            return super.findBeanDeserializer(type, config, beanDescription);
        }
    }

    private boolean isHexadecimalMessageType(final Class<?> rawClass) {
        boolean found = false;

        for (final HexadecimalMessageType hexadecimalMessageType : HexadecimalMessageType.values()) {
            final Class<? extends Message> messageType = hexadecimalMessageType.getMessageType();
            if (messageType.isAssignableFrom(rawClass)) {
                found = true;
                break;
            }
        }

        return found;
    }
}
